// Author: Clint Doriot
// Modified from www.bdmirco.com

#include <avr/io.h>
#include <stdio.h>


////////////////////////////////////////////////////////////////////////////////
// name:    adc_init()
// purpose: Initialize A/D converter to free running, start conversion, use
//            internal 5.0V reference, pre-scale ADC clock to 125 kHz (assuming
//            16 MHz MCU clock)
////////////////////////////////////////////////////////////////////////////////
void adc_init(void)
{
  // Configure PORTF for input
  DDRF  = 0x00;
  PORTF = 0x00;

  // Use AVCC - internal 5.0V reference voltage
  ADMUX = _BV(REFS0);
  
  // Respectively, Enable ADC, Start first conversion, select free running,
  //   use prescalar of 128
  ADCSR = _BV(ADEN)|_BV(ADSC)|_BV(ADFR) | _BV(ADPS2)|_BV(ADPS1)|_BV(ADPS0);
}


////////////////////////////////////////////////////////////////////////////////
// name:    adc_chsel()
// input:   uint8_t channel - the ADC channel to read from
// purpose: Select the specified A/D channel for the next conversion
//            The ATmega128 does not change channels until after the current
//            conversion is complete.
////////////////////////////////////////////////////////////////////////////////
void adc_chsel(uint8_t channel)
{
  // Clears all but first three bits, then sets the channel
  ADMUX = (ADMUX & 0xe0) | (channel & 0x07);
}


////////////////////////////////////////////////////////////////////////////////
// name:    adc_wait()
// purpose: Wait for conversion complete.
////////////////////////////////////////////////////////////////////////////////
void adc_wait(void)
{
  // Checks the interupt flag to see if conversion has completed.
  while ((ADCSR & _BV(ADIF)) == 0)
    ;
}


////////////////////////////////////////////////////////////////////////////////
// name:    adc_start()
// purpose: Start an A/D conversion on the selected channel
////////////////////////////////////////////////////////////////////////////////
void adc_start(void)
{
  // clear conversion, start another conversion
  ADCSR |= _BV(ADIF);
}



////////////////////////////////////////////////////////////////////////////////
// name:    adc_read()
// purpose: Read the currently selected A/D Converter channel.
////////////////////////////////////////////////////////////////////////////////
uint16_t adc_read(void)
{
  return ADC;
}


////////////////////////////////////////////////////////////////////////////////
// name:    adc_readn()
// input:   uint8_t channel - the ADC channel to read from
//          uint8_t n - number of times to sample the ADC
// output:  uint16_t - the average of the n samples
// purpose: Read the specified A/D channel 'n' times and return the average of
//						the samples
////////////////////////////////////////////////////////////////////////////////
uint16_t adc_readn(uint8_t channel, uint8_t n)
{
  uint16_t t;
  uint8_t i;

	// Change channel, reset ADC and wait for first conversion to complete
  adc_chsel(channel);
  adc_start();
  adc_wait();

  adc_start();

  // sample selected channel n times, take the average
  t = 0;
  for (i=0; i<n; i++) {
    adc_wait();
    t += adc_read();
    adc_start();
  }

  // return the average of n samples
  return t / n;
}



