// Author:  Clint Doriot
// Date:    June 8th, 2005
// Updated: July 13th, 2005
//
// File:    lcd_test.c
// Purpose: Create and test a simple interface from an Atmel microcontroller to
//            a Hitachi based LCD screen.
// Notes:   The code is not efficient in the following ways: (1) Uses excess
//						delays to allow for code reusability with other Hitachi LCD models
//						(2) Uses polling process for creating timed delays.


/*----- Includes -------------------------------------------------------------*/
#include <inttypes.h>
#include <avr/io.h>
#include <avr/signal.h>
#include <avr/interrupt.h>
#include "LCD.h"
#include "utils.h"


/*----- Pin Masks ------------------------------------------------------------*/
#define EN 0x10
#define RS 0x20

/*----- LCD VALUES -----------------------------------------------------------*/
const unsigned char CLEAR = 0x01;
const unsigned char CURSOR_SET = 0x80;
const unsigned char LINE1 = 0x00;
const unsigned char LINE2 = 0x40;
const unsigned char LINE3 = 0x14;
const unsigned char LINE4 = 0x54;

////////////////////////////////////////////////////////////////////////////////
// Name:   writeCommand
// Input:  cmd - 8 bit command to be sent to LCD device
// Output: None
// Action: sends a command to an lcd device on PORTA in 4-bit mode (used 2 nibbles)
////////////////////////////////////////////////////////////////////////////////
void writeCommand(unsigned char cmd)
{
  // Create   mask used to affect only pins 0-5 of PORTA
	unsigned char mask;

	// Send First Nybble of Character
	mask = (cmd >> 4) | EN;			 // Set mask for enable and first nybble
	mask &= ~(0xC0);             // clear bits 6,7 so they aren't affected by OR
	PORTA |= mask;               // OR Port and mask, so 1's are written to port
	mask |= 0xC0;                // set bits 6,7 so they aren't affected by AND
	PORTA &= mask;               // AND port and mask, so 0's are written to port
	delay_1ms(5);
	PORTA &= (~EN);              // Clear the Enable pin
	delay_1ms(5);

	// Send Second Nybble of Character
	mask = (cmd & 0x0F) | EN; 	 // Set mask for enable and first nybble
	mask &= ~(0xC0);             // clear bits 6,7 so they aren't affected by OR
	PORTA |= mask;               // OR Port and mask, so 1's are written to port
	mask |= 0xC0;                // set bits 6,7 so they aren't affected by AND
	PORTA &= mask;               // AND port and mask, so 0's are written to port
	delay_1ms(5);
	PORTA &= (~EN);              // Clear the Enable pin
	delay_1ms(5);
}

////////////////////////////////////////////////////////////////////////////////
// Name:   writeChar
// Input:  ch - 8 bit command to be sent to LCD device
// Output: None
// Action: sends a character to an lcd device on PORTA in 4-bit mode (used 2 nibbles)
////////////////////////////////////////////////////////////////////////////////
void writeChar(unsigned char ch)
{
	// Create   mask used to affect only pins 0-5 of PORTA
	unsigned char mask;
	
	// Send First Nybble of Character
	mask = (ch >> 4) | EN | RS;  // Set mask for enable, RS, and first nybble
	mask &= ~(0xC0);             // clear bits 6,7 so they aren't affected by OR
	PORTA |= mask;               // OR Port and mask, so 1's are written to port
	mask |= 0xC0;                // set bits 6,7 so they aren't affected by AND
	PORTA &= mask;               // AND port and mask, so 0's are written to port
	delay_1ms(5);
	PORTA &= (~EN);              // Clear the Enable pin
	delay_1ms(5);
	
	// Send Second Nybble of Character
	mask = (ch & 0x0F) | EN | RS;  // Set mask for enable, RS, and first nybble
	mask &= ~(0xC0);             // clear bits 6,7 so they aren't affected by OR
	PORTA |= mask;               // OR Port and mask, so 1's are written to port
	mask |= 0xC0;                // set bits 6,7 so they aren't affected by AND
	PORTA &= mask;               // AND port and mask, so 0's are written to port
	delay_1ms(5);
	PORTA &= (~EN);              // Clear the Enable pin
	delay_1ms(5);
}

////////////////////////////////////////////////////////////////////////////////
// Name:   initializeLCD
// Input:  None
// Output: None
// Action: Sets up the LCD to accept data properly
////////////////////////////////////////////////////////////////////////////////
void initializeLCD(void)
{
  /*
  Hardware Setup on PortA
    	Purpose:      X  X  RS E  D[7..4]
    	              -  -   - -  - - - -
			Pin No: [msb] 7  6   5 4  3 1 2 0[lsb]
  */

  // Setup PORTA for Output
  DDRA |= 0x3F;

  // Delay 15ms for LCD to Power On
  delay_1ms(15);

  // Start Up Sequence
  writeCommand(0x33);
  writeCommand(0x32); // 0011 00x0: '0011 0010' sets the LCD to 4-bit mode

  // Set up LCD for current purposes
  writeCommand(0x28); // 001x xx00: '0010 1000' sets 4bit mode, 2 lines,
											//	5x7 Character Font (Interface Length)
  writeCommand(0x10); // 0001 xx00: '0001 0000' disables Display Shift,
											//	(which is then set to left shift)
  writeCommand(0x01); // 0000 0001: clears the display
  writeCommand(0x06); // 0000 01xx: '0000 0110' cursor increments after each
											//	byte, doesn't shift
  writeCommand(0x0F); // 0000 1xxx: '0000 1111' turns display on, cursor on,
											//	cursor blinking off
}

////////////////////////////////////////////////////////////////////////////////
// Name:   writeString
// Input:  str[] - character array to be written to the screen
// Output: None
// Action: sends a string to an LCD device
////////////////////////////////////////////////////////////////////////////////
void writeString(unsigned char str[])
{
  int i;
  for(i=0; str[i] != 0; i++)
    writeChar(str[i]);
}

////////////////////////////////////////////////////////////////////////////////
// Name:   writeInt
// Input:  num - number to be written to the screen
// Output: None
// Action: sends an integer to an LCD device
////////////////////////////////////////////////////////////////////////////////
void writeInt(int num, int places)
{
	// Print negative sign and decrease number of available spaces
	if( num < 0)
	{
		writeChar(0xB0);
		places --;
		num = -num;
	}
	
	// Print each digit individually from most significant to least
	int convertZeros = 1;
	for(; places > 0; places--)
	{
	  // Pull out most significant digit and convert to an ASCII character
	  int currentInt = (num / power(10,(places-1)));
		unsigned char currentChar = currentInt + 0x30;
		
		// Convert initial zeroes to spaces
		if(convertZeros && currentChar == 0x30 && places!=1)
			currentChar = 0x20;
		else convertZeros = 0;
		
		// Print current integer as a character
		writeChar(currentChar);

		// Remove most significant digit
	  num -= currentInt*power(10,(places-1));
	}

}

////////////////////////////////////////////////////////////////////////////////
// Name:   setPos
// Input:  int, int - the line and space respectively
// Output: None
// Action: sets the position of the LCD's cursor to a specific spot
////////////////////////////////////////////////////////////////////////////////
void setPos(int row, int col)
{
	// Create basic cursor position command that will be used to set the position
	unsigned char cursorPosCmd = 0x00;

	// Change the command to account for the specified row
	if (row == 1) cursorPosCmd = LINE1;
	else if (row == 2) cursorPosCmd = LINE2;
	else if (row == 3) cursorPosCmd = LINE3;
	else cursorPosCmd = LINE4;
	
	// Change the command to account for the column by offsetting the address
	cursorPosCmd += col;
	
	// Change the command to be Write Cursor Position command by
	cursorPosCmd |= CURSOR_SET;
	
	// Change the cursor
	writeCommand(cursorPosCmd);
	
}

////////////////////////////////////////////////////////////////////////////////
// Name:    clearLCD
// Purpose: clears the entire LCD Screen
////////////////////////////////////////////////////////////////////////////////
void clearLCD(void)
{
	writeCommand(CLEAR);
	setPos(1,0);
}
