// Author: Clint Doriot
// Last Modified: July 21th, 2005

/*----- Includes -------------------------------------------------------------*/
#include <inttypes.h>
#include <avr/io.h>
#include <avr/signal.h>
#include <avr/interrupt.h>
#include "LineTracking.h"
#include "utils.h"
#include "ADC.h"
#include "LCD.h"
#include "servo.h"

#define limit  250

/*----- States ---------------------------------------------------------------*/
#define LFD_nothing   0x00
#define LFD_onLine    0x01
#define LFD_strayedR  0x02
#define LFD_strayedL  0x03
#define inDoor        0x04
#define pastDoor      0x05
#define LFI_nothing   0x06
#define LFI_onLine    0x07
#define LFI_strayedR  0x08
#define LFI_strayedL  0x09
#define found000      0x0A
#define found001      0x0B
#define found010      0x0C
#define found011      0x0D
#define found100      0x0E
#define found101      0x0F
#define found110      0x10
#define found111      0x11
#define finishedMaze  0x12
#define error         0xFF

/*---- Speeds ----------------------------------------------------------------*/
#define LFast 	 1550
#define RFast 	 1560
#define LMed  	 1525
#define RMed  	 1531
#define LSlow 	 1510
#define RSlow 	 1514
#define Stop  	 1500
#define LTurnFor 1550
#define LTurnRev 1450
#define RTurnFor 1575
#define RTurnRev 1425

/*----- Delays ---------------------------------------------------------------*/
// maximum value is 256
#define forwardDelay      150
#define turnDelay    			100
#define intersectionDelay 50

/*----- Line Tracking Sensors ------------------------------------------------*/
uint8_t FR;
uint8_t LO;
uint8_t LM;
uint8_t LI;
uint8_t RI;
uint8_t RM;
uint8_t RO;


////////////////////////////////////////////////////////////////////////////////
// name:    trackLine
// input:   uint8_t currentState - the current state of the line tracker
// outpu:   uint8_t - the new state of the line tracker
// purpose: turns the robot to the left until it is on the new line
////////////////////////////////////////////////////////////////////////////////
uint8_t trackLine(uint8_t currentState)
{
	updateLT();
	
	switch(currentState)
	{
	  case LFD_nothing:
	  case LFD_onLine:
		case LFD_strayedR:
		case LFD_strayedL:
			if ((LI && RI) && !(LM || RM))
				return LFD_onLine;  		// on line, not in doorway
			else if ((LI && RI) && (LM && RM))
			  return inDoor;      		// on line, and in doorway
			else if((LI && !RI) || (LM && LI && RI && !RM))
			  return LFD_strayedR;		// right side not on track properly
			else if((!LI && RI) || (!LM && LI && RI && RM))
			  return LFD_strayedL;    // left side not on track properly
			else
  			return LFD_nothing;
			break;
		case inDoor:
		  //delay_100ms(forwardDelay);
			//updateLT();
		  if(LM || RM)
		    return inDoor;
		  else
		    return pastDoor;
		case pastDoor:
      if ((LI && RI) && !(LM || RM))
		 		return LFI_onLine;    // on line, not in doorway
			else if ((LI && RI) && (LM && RM))
				return inDoor;
			else if((LI && !RI) || (LM && LI && RI && !RM))
				return LFI_strayedR; 		// right side not on track properly
			else if((!LI && RI) || (!LM && LI && RI && RM))
				return LFI_strayedL; 		// left side not on track properly
			else
				return LFI_nothing;
			break;
		case LFI_nothing:
		case LFI_onLine:
		case LFI_strayedR:
		case LFI_strayedL:
		case found000:
		case found001:
		case found010:
		case found011:
		case found100:
		case found101:
		case found110:
		case found111:
			if(LI && RI)
			{
				if(FR && !LO && !LM && !RM && !LM) return LFI_onLine;
				else if(!LO && !LM && !FR && !RM && !RO) return found000;
				else if(!LO && !LM && !FR &&  RM &&  RO) return found001;
				else if(!LO &&  LM &&  FR &&  RM && !RO) return found010;
				else if(!LO && !LM &&  FR &&  RM &&  RO) return found011;
				else if( LO &&  LM && !FR && !RM && !RO) return found100;
				else if( LO &&  LM && !FR &&  RM &&  RO) return found101;
				else if( LO &&  LM &&  FR && !RM && !RO) return found110;
				else if( LO &&  LM &&  FR &&  RM &&  RO) return found111;
				else if((LO + LM)>(RM + RO)) return LFI_strayedR;
				else if((LO + LM)<(RM + RO)) return LFI_strayedL;
				else return LFI_onLine;
			}
			else if(LI && !RI)
				return LFI_strayedR;      // right side not on track properly
			else if(!LI && RI)
				return LFI_strayedL;      // left side not on track properly
			else
  			return LFI_nothing;
			break;
		case finishedMaze:
		  return finishedMaze;
		  break;
		case error:
	  default:
	    return error;
	}
	return error;
}

////////////////////////////////////////////////////////////////////////////////
// name:    updateLineSpeed
// input:   uint8_t currentState -  the state of the line tracker
// purpose: changes the speed of the servos dependent on the current state
////////////////////////////////////////////////////////////////////////////////
void updateLineSpeed(uint8_t currentState)
{
  switch(currentState)
	{
		case LFD_nothing:
		  setRight(Stop);
		  setLeft(Stop);
    	break;
		case LFD_onLine:
		  setRight(RFast);
		  setLeft(LFast);
    	break;
		case LFD_strayedR:
		  setRight(RFast);
		  setLeft(Stop);
    	break;
		case LFD_strayedL:
		  setRight(Stop);
		  setLeft(LFast);
    	break;
		case inDoor:
		  setRight(RFast);
		  setLeft(LFast);
    	break;
		case pastDoor:
		  setRight(RFast);
		  setLeft(LFast);
    	break;
		case LFI_nothing:
		  setRight(Stop);
		  setLeft(Stop);
    	break;
		case LFI_onLine:
		  setRight(RFast);
		  setLeft(LFast);
    	break;
		case LFI_strayedR:
		  setRight(RFast);
		  setLeft(Stop);
    	break;
		case LFI_strayedL:
		  setRight(Stop);
		  setLeft(LFast);
    	break;
		case found000:
		case found001:
		case found010:
		case found011:
		case found100:
		case found101:
		case found110:
		case found111:
		case error:
		  setRight(Stop);
		  setLeft(Stop);
    	break;
		case finishedMaze:
		  setLeft(Stop);
		  setRight(Stop);
		  break;
		default: break;
		}
}


////////////////////////////////////////////////////////////////////////////////
// name:    printLineStatus
// input:   uint8_t currentState -  the state of the line tracker
// purpose: displays the status of the line tracker on the LCD Screen
////////////////////////////////////////////////////////////////////////////////
void printLineStatus(uint8_t currentState)
{
	setPos(1,0);
  switch(currentState)
	{
		case LFD_nothing:
	 		writeString("LFD, I'm Lost!!     "); break;
		case LFD_onLine:
      writeString("LFD, on the line    "); break;
		case LFD_strayedR:
			writeString("LFD, strayed right  "); break;
		case LFD_strayedL:
			writeString("LFD, strayed left   "); break;
		case inDoor:
      writeString("Entered Doorway     "); break;
		case pastDoor:
		  writeString("Entered A New Room  "); break;
		case LFI_nothing:
		  writeString("LFI, I'm Lost!!     "); break;
		case LFI_onLine:
			writeString("LFI, on the line    ");break;
		case LFI_strayedR:
    	writeString("LFI, strayed right  "); break;
		case LFI_strayedL:
		  writeString("LFI, strayed left   "); break;
		case found000:
		 	writeString("Dead End!           ");
			setPos(3,15);
			writeString("000");
			break;
		case found001:
		  writeString("Can Only Turn Right ");
			setPos(3,15);
			writeString("001");
			break;
		case found010:
		  writeString("Can Only Go Forward ");
			setPos(3,15);
			writeString("010");
			break;
		case found011:
		  writeString("Can Go Forward,Right");
			setPos(3,15);
			writeString("011");
			break;
		case found100:
		  writeString("Can Only Go Left    ");
			setPos(3,15);
			writeString("100");
			break;
		case found101:
		  writeString("Can Go Left, Right  ");
			setPos(3,15);
			writeString("101");
			break;
		case found110:
		  writeString("Can Go Left, Forward");
			setPos(3,15);
			writeString("110");
			break;
		case found111:
		  writeString("Can Go Any Direction");
			setPos(3,15);
			writeString("111");
			break;
		case finishedMaze:
		  writeString("Finished The Maze   "); break;
		case error:
		default: break;
		}
}




////////////////////////////////////////////////////////////////////////////////
// name:    updateLT
// purpose: updates the values of all the Line Tracking Sensors and prints them
////////////////////////////////////////////////////////////////////////////////
void updateLT(void)
{
	
	// Check each line sensor and edit LT_PORT if they are on a line
	if(adc_readn(0, 5)<limit) FR = 1;
		else FR = 0;
	if(adc_readn(1, 5)<limit) RO = 1;
		else RO = 0;
	if(adc_readn(2, 5)<limit) RM = 1;
		else RM = 0;
	if(adc_readn(3, 5)<limit) RI = 1;
		else RI = 0;
	if(adc_readn(4, 5)<limit) LI = 1;
		else LI = 0;
	if(adc_readn(5, 5)<limit) LM = 1;
		else LM = 0;
	if(adc_readn(6, 5)<limit) LO = 1;
		else LO = 0;
}

////////////////////////////////////////////////////////////////////////////////
// name:    centerWheels
// purpose: moves the robot forward to center the axis over the intersection
////////////////////////////////////////////////////////////////////////////////
void centerWheels(void)
{
	// Center the wheels on the intersection
	setRight(RMed);
	setLeft(LMed);
	
	setPos(1,0);
	writeString("Centering Wheels    ");
	
	delay_100ms(forwardDelay);

	setRight(Stop);
	setLeft(Stop);
}


////////////////////////////////////////////////////////////////////////////////
// name:    turnRight
// purpose: turns the robot to the right until it is on the new line
////////////////////////////////////////////////////////////////////////////////
void turnRight(void)
{
  centerWheels();

	setRight(RTurnRev);
	setLeft(LTurnFor);

	setPos(1,0);
	writeString("Turning Right       ");

	while((adc_readn(4, 5)<limit) || (adc_readn(3, 5)<limit) ||
	      (adc_readn(2, 5)<limit) || (adc_readn(5, 5)<limit))
	{
	  //wait for middle sensors to be off the line
	}
	while((adc_readn(4, 5)>limit) || (adc_readn(3, 5)>limit))
	{
	  //wait for middle sensors to be on the line
	}

	setRight(Stop);
	setLeft(Stop);
}

////////////////////////////////////////////////////////////////////////////////
// name:    turnLeft
// purpose: turns the robot to the left until it is on the new line
////////////////////////////////////////////////////////////////////////////////
void turnLeft(void)
{
	centerWheels();

	setRight(RTurnFor);
	setLeft(LTurnRev);

	setPos(1,0);
	writeString("Turning Left        ");
	
  while((adc_readn(4, 5)<limit) || (adc_readn(3, 5)<limit) ||
	      (adc_readn(2, 5)<limit) || (adc_readn(5, 5)<limit))
	{
	  //wait for middle sensors to be off the line
	}
	while((adc_readn(4, 5)>limit) || (adc_readn(3, 5)>limit))
	{
	  //wait for middle sensors to be on the line
	}
	
	setRight(Stop);
	setLeft(Stop);
}


////////////////////////////////////////////////////////////////////////////////
// name:    turnAround
// purpose: turns the robot to the right until it has turned around and it on
//          the new line.
////////////////////////////////////////////////////////////////////////////////
void turnAround(void)
{
  centerWheels();

	setRight(RTurnFor);
	setLeft(LTurnRev);

	setPos(1,0);
	writeString("Turning Around      ");
	
 while((adc_readn(4, 5)<limit) || (adc_readn(3, 5)<limit) ||
	      (adc_readn(2, 5)<limit) || (adc_readn(5, 5)<limit))
	{
	  //wait for middle sensors to be off the line
	}
	while((adc_readn(4, 5)>limit) || (adc_readn(3, 5)>limit))
	{
	  //wait for middle sensors to be on the line
	}
	
	//Do it the second time to get past the first turn marker
	while((adc_readn(4, 5)<limit) || (adc_readn(3, 5)<limit) ||
	      (adc_readn(2, 5)<limit) || (adc_readn(5, 5)<limit))
	{
	  //wait for middle sensors to be off the line
	}
	while((adc_readn(4, 5)>limit) || (adc_readn(3, 5)>limit))
	{
	  //wait for middle sensors to be on the line
	}
	
	setRight(Stop);
	setLeft(Stop);
}

////////////////////////////////////////////////////////////////////////////////
// name:    debounceIntersection
// input:   uint8_t currentState - the currentState of the line tracker sensors
// output:  uint8_t - the modified currentState of the line tracker sensors
// purpose: ensures that the line following sensors have read the intersection
//          properly by moving forward just slightly to ensure all sensors
//          are on the lines correctly
////////////////////////////////////////////////////////////////////////////////
uint8_t debounceIntersection(uint8_t currentState)
{
	if(currentState >=found000 && currentState <= found111)
	{
		setPos(1,0);
		writeString("DebounceIntersection");
		
		setLeft(LMed);
		setRight(RMed);
		delay_100ms(intersectionDelay);
		
  	currentState = trackLine(currentState);
	}

	return currentState;
}
