// Author: Clint Doriot
// Last Modified: July 20th, 2005

/*----- Includes -------------------------------------------------------------*/
#include <inttypes.h>
#include <avr/io.h>
#include <avr/signal.h>
#include <avr/interrupt.h>
#include "LineTracking.h"
#include "utils.h"
#include "ADC.h"
#include "LCD.h"
#include "servo.h"
#include "MazeSolving.h"

#define NULL 0

/*----- Globals --------------------------------------------------------------*/
struct node squares[6][6];
uint8_t orientation;
uint8_t currentX, currentY;
uint8_t targetX, targetY;
uint8_t targetHasChanged, infoHasChanged;
uint8_t path[36];
uint8_t pathPos;
uint8_t finishes;

/*----- Orientations ---------------------------------------------------------*/
#define UP      0x00
#define RIGHT   0x01
#define DOWN    0x02
#define LEFT    0x03
#define UNUSED  0x04

/*----- States ---------------------------------------------------------------*/
#define LFD_nothing   0x00
#define LFD_onLine    0x01
#define LFD_strayedR  0x02
#define LFD_strayedL  0x03
#define inDoor        0x04
#define pastDoor      0x05
#define LFI_nothing   0x06
#define LFI_onLine    0x07
#define LFI_strayedR  0x08
#define LFI_strayedL  0x09
#define found000      0x0A
#define found001      0x0B
#define found010      0x0C
#define found011      0x0D
#define found100      0x0E
#define found101      0x0F
#define found110      0x10
#define found111      0x11
#define finishedMaze  0x12
#define error         0xFF


////////////////////////////////////////////////////////////////////////////////
// name:    initializeMazeSolver
// purpose: Sets up the initial values of the maze and the data structures used
//            to represent each node
////////////////////////////////////////////////////////////////////////////////
void initializeMazeSolver(void)
{
	//setPos(1,0);
	//writeString("Initialization");
	//delay_100ms(250);
	
	// Initialize location information
	currentX = 0;
	currentY = 5;
	orientation = UP;
	printLocation();
	printOrientation();
	
	targetX = 2;
	targetY = 1;
	finishes = 0;
	
	targetHasChanged = 0;
	infoHasChanged = 1;
	
	// initialize path from current square to destination
	int i=0;
	for(i=0; i < 36; i++)
	{
	  path[i] = UNUSED;
	}
	
	// initialize each node
	int x=0, y=0;
	for(x=0; x<6; x++)
	  for(y=0; y<6; y++)
	  {
	    squares[x][y].closed = 0;
	    squares[x][y].open = 0;
	    squares[x][y].visited = 0;
	    squares[x][y].posX = x;
	    squares[x][y].posY = y;
	    squares[x][y].G = 0;
	    squares[x][y].H = 0;

			if(y!=0)
			  squares[x][y].borders[UP] = &squares[x][y-1];
			else squares[x][y].borders[UP] = NULL;
			
			if(x!=5)
			  squares[x][y].borders[RIGHT] = &squares[x+1][y];
			else squares[x][y].borders[RIGHT] = NULL;
			
			if(y!=5)
			  squares[x][y].borders[DOWN] = &squares[x][y+1];
			else squares[x][y].borders[DOWN] = NULL;
			
			if(x!=0)
			  squares[x][y].borders[LEFT] = &squares[x-1][y];
			else squares[x][y].borders[LEFT] = NULL;
	  }
	  
	  // initialize starting node
	  squares[currentX][currentY].borders[LEFT] = NULL;
	  squares[currentX][currentY].borders[RIGHT] = NULL;
	  squares[currentX][currentY].borders[DOWN] = NULL;
    squares[currentX][currentY].visited = 1;
	  squares[currentX][currentY].G = 0;
	  heuristic(currentX, currentY);
	  currentY--;
	  
	  solvePath();
}


////////////////////////////////////////////////////////////////////////////////
// name:    updateIntersection
// input:   uint8_t intersection - the type of intersection the robot has
//            just encountered.
// purpose: if the current node has not been visited before, this function
//            updates the current node with the newly gained information
////////////////////////////////////////////////////////////////////////////////
void updateIntersection(uint8_t intersection)
{
	if(!squares[currentX][currentY].visited)
	{
	  infoHasChanged = 1;
	  squares[currentX][currentY].visited = 1;
	  struct node * current = &squares[currentX][currentY];
	  
		switch(intersection)
		{
		  case found000:
				// Delete Links to this Square
				if(current->borders[(orientation+LEFT)%4] != NULL)
				  current->borders[(orientation+LEFT)%4]->borders[(orientation+LEFT+2)%4] = NULL;
				if(current->borders[(orientation+UP)%4] != NULL)
				  current->borders[(orientation+UP)%4]->borders[(orientation+UP+2)%4] = NULL;
				if(current->borders[(orientation+RIGHT)%4] != NULL)
				  current->borders[(orientation+RIGHT)%4]->borders[(orientation+RIGHT+2)%4] = NULL;
		  
		    // Delete Links from this Square
		    current->borders[(orientation+LEFT)%4] = NULL;
		    current->borders[(orientation+UP)%4] = NULL;
		    current->borders[(orientation+RIGHT)%4] = NULL;
				//writeString("000");
		    break;
      case found001:
				// Delete Links to this Square
				if(current->borders[(orientation+LEFT)%4] != NULL)
				  current->borders[(orientation+LEFT)%4]->borders[(orientation+LEFT+2)%4] = NULL;
				if(current->borders[(orientation+UP)%4] != NULL)
				  current->borders[(orientation+UP)%4]->borders[(orientation+UP+2)%4] = NULL;

		    // Delete Links from this Square
		    current->borders[(orientation+LEFT)%4] = NULL;
		    current->borders[(orientation+UP)%4] = NULL;
		    //writeString("001");
		    break;
      case found010:
				// Delete Links to this Square
				if(current->borders[(orientation+LEFT)%4] != NULL)
				  current->borders[(orientation+LEFT)%4]->borders[(orientation+LEFT+2)%4] = NULL;
				if(current->borders[(orientation+RIGHT)%4] != NULL)
				  current->borders[(orientation+RIGHT)%4]->borders[(orientation+RIGHT+2)%4] = NULL;

		    // Delete Links from this Square
		    current->borders[(orientation+LEFT)%4] = NULL;
		    current->borders[(orientation+RIGHT)%4] = NULL;
		    //writeString("010");
		    break;
      case found011:
				// Delete Links to the Square
				if(current->borders[(orientation+LEFT)%4] != NULL)
				  current->borders[(orientation+LEFT)%4]->borders[(orientation+LEFT+2)%4] = NULL;

		    // Delete Links from this Square
		    current->borders[(orientation+LEFT)%4] = NULL;
		    //writeString("011");
		    break;
      case found100:
        // Delete Links to the Square
				if(current->borders[(orientation+UP)%4] != NULL)
				  current->borders[(orientation+UP)%4]->borders[(orientation+UP+2)%4] = NULL;
				if(current->borders[(orientation+RIGHT)%4] != NULL)
				  current->borders[(orientation+RIGHT)%4]->borders[(orientation+RIGHT+2)%4] = NULL;
        
        // Delete Links from the Square
		    current->borders[(orientation+UP)%4] = NULL;
		    current->borders[(orientation+RIGHT)%4] = NULL;
		    //writeString("100");
		    break;
      case found101:
        // Delete Links to the Square
				if(current->borders[(orientation+UP)%4] != NULL)
				  current->borders[(orientation+UP)%4]->borders[(orientation+UP+2)%4] = NULL;

        // Delete Links from the Square
		    current->borders[(orientation+UP)%4] = NULL;
		    //writeString("101");
		    break;
      case found110:
        // Delete Links to the Square
				if(current->borders[(orientation+RIGHT)%4] != NULL)
				  current->borders[(orientation+RIGHT)%4]->borders[(orientation+RIGHT+2)%4] = NULL;

        // Delete Links from the Square
		    current->borders[(orientation+RIGHT)%4] = NULL;
		    //writeString("110");
		    break;
      case found111:
        // keep all current nodes
        //writeString("111");
		    break;
		}
	}
}


////////////////////////////////////////////////////////////////////////////////
// name:    getNextMove
// output:  uint8_t - represents the type of move to make
// purpose: returns the next move that the robot should take in order to
//            complete the path from the current location to the target
////////////////////////////////////////////////////////////////////////////////
uint8_t getNextMove(void)
{
	// Recalculate Path if maze information has changed
	if(infoHasChanged || targetHasChanged)
	  solvePath();

	// Get Next Move from end of array to process it
	uint8_t nextMove;
	if(pathPos < 36)
	{
		nextMove = path[pathPos];
		path[pathPos] = UNUSED;
		pathPos--;

		//Temp Error Code
		printNextMove(nextMove);
  }
	else nextMove = UNUSED;

	// Update location to upcoming square
	switch(nextMove)
	{
	  case UP:
	    currentY--;
			break;
		case RIGHT:
		  currentX++;
		  break;
		case DOWN:
		  currentY++;
		  break;
		case LEFT:
		  currentX--;
		  break;
	}
	
	//Convert from absolute position to relative
	nextMove = (4 + nextMove - orientation) % 4;

	return nextMove;

}


////////////////////////////////////////////////////////////////////////////////
// name:    solvePath
// purpose: Solves the maze using the known information (location, target, and
//            node info) and the A* Algorithm. Must be recalculated each time
//            the target or node information changes in order to be effective.
////////////////////////////////////////////////////////////////////////////////
void solvePath(void)
{
	// Algorithm Setup
	uint8_t solvedMaze = 0;
	clearPath();
	uint8_t x = currentX;
	uint8_t y = currentY;
	squares[x][y].open = 1;
	
	// See if maze is solved, or unsolvable
	while(!solvedMaze && (!squares[targetX][targetY].open) && !openListIsEmpty())
	{
		// Get lowest cost square from open list
		struct node * currentNode = lowestOpenNode();
		currentNode->open=0;
		currentNode->closed=1;
		
		int i=0;
		for(i = 0; i<4 && !solvedMaze; i++)
		{
			//Check Upper Square if it exists and isn't on the closed list
			if(currentNode->borders[i]!=NULL && !currentNode->borders[i]->closed)
			{
			
		  	if(!currentNode->borders[i]->open)
		  	{
		    	// if this is the first time its reachable, add it too the open list
		    	// and calculate costs
        	currentNode->borders[i]->open = 1;
        	currentNode->borders[i]->parent = currentNode;
        	currentNode->borders[i]->G = 1 + currentNode->G;
		  	}
		  	else
		  	{
		    	// if already on open list, see if new cost is better
		    	if((currentNode->borders[i]->G) > (1 + currentNode->G))
		    	{
		      	currentNode->borders[i]->parent = currentNode;
        		currentNode->borders[i]->G = 1 + currentNode->G;
		    	}
		  	}
		  	
		  	// check to see if border node is the target
				if((currentNode->borders[i]->posX == targetX) &&
           (currentNode->borders[i]->posY == targetY))
				solvedMaze = 1;
			}
		} //end for loop for directions
	} //end while loop for solving maze
	
	if(!solvedMaze)
	{
	  //setPos(1,0);
	  //writeString("Error: No Soln");
	}
	else
	{
	  // Set Path from Finish to Start
	  pathPos = 0;
		struct node * currentNode = &squares[targetX][targetY];

	  //Here, UP, DOWN, RIGHT, LEFT, are absolute terms with respect to board
		while(!((currentNode->posX == currentX) && (currentNode->posY == currentY)))
		{
			if(currentNode->posX > currentNode->parent->posX)
			{
				path[pathPos] = RIGHT;
			}
			else if(currentNode->posX < currentNode->parent->posX)
			{
				path[pathPos] = LEFT;
			}
			else if(currentNode->posY > currentNode->parent->posY)
			{
				path[pathPos] = DOWN;
			}
			else if(currentNode->posY < currentNode->parent->posY)
			{
				path[pathPos] = UP;
			}
			else
			{
				path[pathPos] = UNUSED;
			}
				
			currentNode = currentNode->parent;
			pathPos++;
		}
  }
  pathPos--;
  printPath();
}


////////////////////////////////////////////////////////////////////////////////
// name:    lowestOpenNode
// output:  *node - a pointer to the node with the lowest F value
// purpose: finds and returns the node with the lowest F value. This process is
//            inefficient for large maps because it checks every single node on
//            the map.
////////////////////////////////////////////////////////////////////////////////
struct node *lowestOpenNode(void)
{
	int x,y;
	struct node * lowest = NULL;
	
	for(x=0; x<6; x++)
	  for(y=0; y<6; y++)
	  {
	    struct node * current = &squares[x][y];
	    if(current!=NULL && current->open &&
			   (lowest==NULL || (current->G + current->H) < (lowest->G + lowest->H)))
				lowest = current;
		}
		
	return lowest;
}


////////////////////////////////////////////////////////////////////////////////
// name:    openListIsEmpty
// output:  uint8_t - returns 0 if the open list is not empty, 1 if it is empty
// purpose: returns 0 if the open list is not empty, 1 if it is empty
////////////////////////////////////////////////////////////////////////////////
uint8_t openListIsEmpty(void)
{
	struct node * lowest = lowestOpenNode();
	if (lowest == NULL)
		return 1;
	else return 0;
}


////////////////////////////////////////////////////////////////////////////////
// name:    clearPath
// purpose: resets all basic information used to calculated path between the
//            start and finish. It does not reset the information on a square's
//            connected nodes
////////////////////////////////////////////////////////////////////////////////
void clearPath(void)
{
  targetHasChanged = 0;
	infoHasChanged = 0;
	
	// reset path from current square to destination
	int i=0;
	for(i=0; i < 36; i++)
	{
	  path[i] = UNUSED;
	}
	pathPos = 0;
	
	int x;
	int y;
	
	for(x=0; x<6; x++)
	  for(y=0; y<6; y++)
	  {
	    // reset closed list, open list, G, H, and parent
	    squares[x][y].closed=0;
	    squares[x][y].open=0;
	    squares[x][y].G=0;
	    squares[x][y].H=0;
	    squares[x][y].parent=NULL;
	  }
	setPos(2,0);
	writeString("                    ");

}


////////////////////////////////////////////////////////////////////////////////
// name:    heuristic
// input:   uint8_t x - the x position of the node to check
// input:   uint8_t y - the y position of the node to check
// purpose: Calculates the estimated cost of going from the current square to
//            the target square. Uses the "Manhatten Method" to estimate this
//            cost. The closer the estimate is to the actual cost, the quicker
//            the algorithm will sovle the maze; however, if the estimate is
//            greater than the actual cost, the algorithm is not guarenteed to
//            find a suitable path.
////////////////////////////////////////////////////////////////////////////////
void heuristic(uint8_t x, uint8_t y)
{
	uint8_t distX = 0;
	uint8_t distY = 0;
	
	// "Manhatten Method"
	if (x > targetX)
	  distX = x - targetX;
	else distX = targetX - x;
	
	if (y > targetY)
	  distY = y - targetY;
	else distY = targetY - y;
	
	squares[x][y].H = distX + distY;
}

////////////////////////////////////////////////////////////////////////////////
// name:    turnRobot
// input:   uint8_t currentState - the current state of the line tracker sensors
// output:  uint8_t - the new state, now that the robot is looking for a door
// purpose: turns the robot dependent on the current orientation and the next
//          step to take.
////////////////////////////////////////////////////////////////////////////////
uint8_t turnRobot(uint8_t currentState)
{

		uint8_t move = 0x05;
		
		// check to see if its at an intersection
		if(currentState == finishedMaze) return finishedMaze;
		
		if(currentState >= found000 && currentState <= found111)
		{
		  if(currentX == targetX && currentY == targetY)
		  {
		    if(finishes==0)
		    {
					targetX = 4;
					targetY = 4;
					targetHasChanged=1;
				}
				else if(finishes==1)
				{
					targetX = 0;
					targetY = 5;
					targetHasChanged=1;
				}
				else if(finishes==2)
				{
					targetX = 2;
					targetY = 1;
					targetHasChanged=1;
				}
				else if(finishes>=3)
				{
				  return finishedMaze;
				}
				  
				finishes++;
			}
		
		  updateIntersection(currentState);
		  move = getNextMove();
		}
		
		// Move Robot Depending on intersection and next move
		// terms DOWN, UP, LEFT, RIGHT are relative to orientation at this part
		switch(currentState)
		{
			case inDoor:
			  printLocation();
			  break;
			case found000:
        if(move==DOWN)
        {
          centerWheels();
          turnRight();
          orientation = (orientation + 2)%4;
					printOrientation();
				}
				currentState = LFD_onLine;
				break;
			case found001:
        if(move==DOWN)
        {
          centerWheels();
          turnLeft();
					orientation = (orientation + 2)%4;
					printOrientation();
				}
				else if(move==RIGHT)
				{
          centerWheels();
				  turnRight();
				  orientation = (orientation + 1)%4;
					printOrientation();
				}
				currentState = LFD_onLine;
				break;
			case found010:
        if(move==DOWN)
        {
          centerWheels();
          turnLeft();
				  orientation = (orientation + 2)%4;
					printOrientation();
				}
				else if(move==UP)
				  centerWheels();
				currentState = LFD_onLine;
				break;
			case found011:
        if(move==DOWN)
        {
          centerWheels();
          turnLeft();
				  orientation = (orientation + 2)%4;
					printOrientation();
				}
				else if(move==RIGHT)
				{
          centerWheels();
				  turnRight();
				  orientation = (orientation + 1)%4;
					printOrientation();
				}
				else if(move==UP)
				  centerWheels();
				currentState = LFD_onLine;
				break;
      case found100:
        if(move==DOWN)
        {
          centerWheels();
          turnRight();
				  orientation = (orientation + 2)%4;
					printOrientation();
				}
				else if(move==LEFT)
				{
          centerWheels();
				  turnLeft();
				  orientation = (orientation + 3)%4;
					printOrientation();
				}
				currentState = LFD_onLine;
				break;
      case found101:
        if(move==DOWN)
        {
          centerWheels();
          turnAround();
				  orientation = (orientation + 2)%4;
					printOrientation();
				}
        else if(move==LEFT)
        {
          centerWheels();
				  turnLeft();
				  orientation = (orientation + 3)%4;
					printOrientation();
				}
				else if(move==RIGHT)
				{
          centerWheels();
				  turnRight();
				  orientation = (orientation + 1)%4;
					printOrientation();
				}
				currentState = LFD_onLine;
				break;
      case found110:
        if(move==DOWN)
        {
          centerWheels();
          turnRight();
				  orientation = (orientation + 2)%4;
					printOrientation();
				}
        else if(move==LEFT)
        {
          centerWheels();
				  turnLeft();
				  orientation = (orientation + 3)%4;
					printOrientation();
				}
				else if(move==UP)
				  centerWheels();
				currentState = LFD_onLine;
				break;
      case found111:
        if(move==DOWN)
        {
          centerWheels();
					turnAround();
				  orientation = (orientation + 2)%4;
					printOrientation();
				}
        else if(move==LEFT)
        {
          centerWheels();
				  turnLeft();
				  orientation = (orientation + 3)%4;
					printOrientation();
				}
				else if(move==RIGHT)
				{
          centerWheels();
				  turnRight();
				  orientation = (orientation + 1)%4;
					printOrientation();
				}
        else if(move==UP)
				  centerWheels();
				currentState = LFD_onLine;
				break;
		}
		
		return currentState;
}


////////////////////////////////////////////////////////////////////////////////
// printing funtions for debugging purposes

void printOrientation(void)
{
	setPos(4,0);
	writeString("O: ");
	switch(orientation)
	{
	  case UP:
	    writeString("UP   ");
	    break;
		case DOWN:
		  writeString("DOWN ");
		  break;
		case LEFT:
		  writeString("LEFT ");
		  break;
		case RIGHT:
		  writeString("RIGHT");
		  break;
		default:
		  writeString("ERROR");
		  break;
	}
}

void printLocation(void)
{
	setPos(4,10);
	writeString("(");
	writeInt(currentX,1);
	writeString(",");
	writeInt(currentY,1);
	writeString(")");
}

void printPath(void)
{
	setPos(2,0);
	int i;
	for(i=pathPos; i>=0; i--)
	{
		switch(path[i])
		{
		  case UP:
		    writeString("U");
		    break;
		  case DOWN:
		    writeString("D");
		    break;
		  case RIGHT:
		    writeString("R");
		    break;
		  case LEFT:
		    writeString("L");
		    break;
		  case UNUSED:
		    writeString("E");
		    break;
		  default:
		    writeString("X");
		    break;
		
		}
	}
}

void printNextMove(uint8_t nextMove)
{
		setPos(3, 0);
		writeString("Next:");
		switch(nextMove)
		{
	  	case UP:
	    	writeString("UP   ");
		    break;
			case DOWN:
			  writeString("DOWN ");
		  	break;
			case LEFT:
			  writeString("LEFT ");
			  break;
			case RIGHT:
			  writeString("RIGHT");
			  break;
			default:
			  writeString("ERROR");
			  break;
		}
}
	
