// Author:   Clint Doriot
// Date:     June 8th, 2005
// Last Mod: July 14th, 2005
//
// File:    servos.c
// Purpose: Simple interface from the Microcontroller to four servos

#include <inttypes.h>
#include <avr/io.h>
#include <avr/signal.h>
#include <avr/interrupt.h>
#include "utils.h"

/*----- Servos -----*/
#define RIGHT 	OCR1C // PORTB pin 7; Servo Header2
#define LEFT 		OCR1B // PORTB pin 6; Servo Header1
#define PAN     OCR3C // PORTE pin 5; Servo Header5
#define TILT    OCR3B // PORTE pin 4; Servo Header4

#define servoDelay 25


////////////////////////////////////////////////////////////////////////////////
// Name:   initializeServos
// Input:  None
// Output: None
// Action: Sets up the Servo Interface
////////////////////////////////////////////////////////////////////////////////
void initializeServos(void)
{
  /*
  Hardware Setup
      The power, ground, and signal are controlled by the Servo Header Pins on
			the BDMicro Board, but these ports require an input signal from one of
			the standard ports, PORTB  */

  DDRB = 0xC0;    // Sets up PORTB for output on pins 7, 6
  TCCR1A = 0xA8;
  TCCR1B = 0x12;
  ICR1 = 20000;
  
  DDRE = 0xFF;    // Sets up PORTE for output on pins 7, 6
  TCCR3A = 0xA8;
  TCCR3B = 0x12;
  ICR3 = 20000;
  
	PAN = 1500;
	TILT = 1500;

}

////////////////////////////////////////////////////////////////////////////////
// Name:   setRight
// Input:  int speed - the new speed, ranging from 1000 to 2000
// Output: None
// Action: Adjusts the speed of the right servo
////////////////////////////////////////////////////////////////////////////////
void setRight(int speed)
{
	if(speed>=1000 && speed<=2000)
	  RIGHT = speed;
}

////////////////////////////////////////////////////////////////////////////////
// Name:   setleft
// Input:  int speed - the new speed, ranging from 1000 to 2000
// Output: None
// Action: Adjusts the speed of the left servo
////////////////////////////////////////////////////////////////////////////////
void setLeft(int speed)
{
	if(speed>=1000 && speed<=2000)
		LEFT = speed;
}

////////////////////////////////////////////////////////////////////////////////
// Name:   setPan
// Input:  int position - the new position, ranging from 1000 to 2000
// Output: None
// Action: Adjusts the position of the pan servo
////////////////////////////////////////////////////////////////////////////////
void setPan(int position)
{
	// ensure new position is within proper limits
 	if(position>=1000 && position<=2000)
	{
	  if(position > PAN)
	  {
	    // if new position is greater than old position
	    int i;
	  	for(i = PAN; i<=position; i++)
      {
	  	  delay_1ms(servoDelay);
	    	PAN = i;
			}
		}
		else if (position < PAN)
		{
		  // if new position is less than old position
		  int i;
		  for(i = PAN; i >=position; i--)
      {
	  	  delay_1ms(servoDelay);
	    	PAN = i;
			}
		}
		else PAN = position;
	}
}

////////////////////////////////////////////////////////////////////////////////
// Name:   setTilt
// Input:  int speed - the new position, ranging from 1000 to 2000
// Output: None
// Action: Adjusts the position of the tilt servo
////////////////////////////////////////////////////////////////////////////////
void setTilt(int position)
{
	// ensure new position is within proper limits
 	if(position>=1000 && position<=2000)
	{
	  if(position > TILT)
	  {
	    // if new position is greater than old position
	    int i;
	  	for(i = TILT; i<=position; i++)
	  	{
	  	  delay_1ms(servoDelay);
	    	TILT = i;
			}
		}
		else if (position < TILT)
		{
		  // if new position is less than old position
		  int i;
		  for(i = TILT; i >=position; i--)
    	{
	  	  delay_1ms(servoDelay);
	    	TILT = i;
			}
		}
		else TILT = position;
	}
}


