// Author: Clint Doriot
// Special thanks to Julio Suarez for helping interface the sonar with portions
// of his code

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/signal.h>
#include <inttypes.h>
#include "sonar.h"
#include "servo.h"
#include "LCD.h"


volatile uint16_t ms_count;
volatile int sonar;
volatile int Tdist;
volatile int dist;
volatile int in;
volatile int sfbusy;
volatile int sonarCount, avg;

#define LOOKING_FOR_OBJECT 0x01
#define LOOKING_FOR_END    0x02

////////////////////////////////////////////////////////////////////////////////
//
//
////////////////////////////////////////////////////////////////////////////////
void initializeSonar(void)
{

	// Set up PortC for using Sonar
	DDRC |= 0x08; // sonar output on pin 3
	DDRC &= 0xFB; // sonar input on pin 2
	PORTC = 0x00;

	sonar = 0;
	dist = 1000;
	sfbusy = 0;

 	//
 	TIFR  |= _BV(OCIE0);
	TCCR0  = _BV(WGM01)|_BV(CS02)|_BV(CS00); /* CTC, prescale = 128 */
  TCNT0  = 0;
  TIMSK |= _BV(OCIE0);    /* enable output compare interrupt */
  OCR0   = 6;          		/* match in 50 us */
}


////////////////////////////////////////////////////////////////////////////////
//
//
////////////////////////////////////////////////////////////////////////////////
void sampleSonar(uint8_t samples)
{
	int i;
	int sum = 0;
	for(i=0; i!= samples; i++)
	{
		if (sfbusy == 0)
		{
			sfbusy = 1;
			triggerSonar();
			sonar = 1;
		}

		while(1)
		{
			if(sfbusy == 0)
				break;
		}

		sum += dist;
		us50_sleep(200);
	}

	dist = sum / samples;
}


////////////////////////////////////////////////////////////////////////////////
//
//
////////////////////////////////////////////////////////////////////////////////
void triggerSonar(void)
{
	PORTC |= 0x08;	//Trigger high
	us50_sleep(1);
	PORTC &= 0xF7;
	us50_sleep(5);
}


////////////////////////////////////////////////////////////////////////////////
//
//
////////////////////////////////////////////////////////////////////////////////
INTERRUPT(SIG_OUTPUT_COMPARE0)
{
  ms_count++;

  // start sonar routine
  if(sonar == 1)
  {
  	in = PINC & (0x04);

	  if(in != 0)
	  	Tdist = Tdist + 1;
		else
	  {
	  	dist= Tdist;
	  	Tdist = 0;
	  	sonar = 0;
	  	sfbusy = 0;
	  }
  }
}


////////////////////////////////////////////////////////////////////////////////
//
//
////////////////////////////////////////////////////////////////////////////////
void us50_sleep(uint16_t us50)
{
  TCNT0  = 0;
  ms_count = 0;
  while (ms_count != us50);
}

////////////////////////////////////////////////////////////////////////////////
// name:    scanArea
// output:  int - the location Pan location that the sonar has found the closed
//          object at
// purpose: rotates the camera/sonar mount in search of the closed object.
//          records the first and last angle this object appeared at and returns
//          the average.
////////////////////////////////////////////////////////////////////////////////
int scanArea(void)
{
	int lookingForEnd = 0;
	int firstPan = 700;
	int lastPan = 700;
	int maxPan = 2300;
	int currentPan = 700;
	int bestDist = 1000;
	
	// Scan entire area
	while(currentPan <= maxPan)
	{
	  // Set Position of Sonar and check distance
	  setPan(currentPan);
   	sampleSonar(10);


		// Check if camera finds a closer object
	  if(dist < (bestDist-10))
	  {
	    // Keep track of best position and distance
	    bestDist = dist;
	    firstPan = currentPan;
	    
	    lookingForEnd = 1;
	  }
	  
	  // Check if camera finds end of current object
	  if(lookingForEnd && (dist > (bestDist+10)))
	  {
	    // Keep track of best position and distance
	    lastPan = currentPan;

	    lookingForEnd = 0;
	  }
	  
	  // Update LCD
   	setPos(1,0);
   	writeString("Position: ");
   	writeInt(currentPan, 5);

   	setPos(2,0);
   	writeInt(firstPan, 6);
   	writeInt((firstPan+lastPan)/2, 6);
   	writeInt(lastPan, 6);

   	setPos(3,0);
   	writeString("Sonar:    ");
   	writeInt(dist, 5);

   	setPos(4,0);
   	writeString("Best:     ");
   	writeInt(bestDist, 5);


		// Increment Position
   	currentPan+=10;
	  
	}

	
	//return ((firstPan + lastPan)/2);
	return (firstPan+lastPan)/2;
	
}
